package record

import (
	"fmt"
	"strconv"
	"time"

	abraFormatter "coopcloud.tech/abra/cli/formatter"
	"coopcloud.tech/abra/cli/internal"
	gandiPkg "coopcloud.tech/abra/pkg/dns/gandi"
	"github.com/libdns/gandi"
	"github.com/libdns/libdns"
	"github.com/sirupsen/logrus"
	"github.com/urfave/cli/v2"
)

// RecordNewCommand creates a new domain name record.
var RecordNewCommand = &cli.Command{
	Name:      "new",
	Usage:     "Create a new domain record",
	Aliases:   []string{"n"},
	ArgsUsage: "<zone>",
	Flags: []cli.Flag{
		internal.DNSProviderFlag,
		internal.DNSTypeFlag,
		internal.DNSNameFlag,
		internal.DNSValueFlag,
		internal.DNSTTLFlag,
		internal.DNSPriorityFlag,
	},
	Description: `
This command creates a new domain name record for a specific zone.

You must specify a zone (e.g. example.com) under which your domain name records
are listed. This zone must already be created on your provider account.

Example:

    abra record new foo.com -p gandi -t A -n myapp -v 192.168.178.44

You may also invoke this command in "wizard" mode and be prompted for input

    abra record new
`,
	Action: func(c *cli.Context) error {
		zone, err := internal.EnsureZoneArgument(c)
		if err != nil {
			logrus.Fatal(err)
		}

		if err := internal.EnsureDNSProvider(); err != nil {
			logrus.Fatal(err)
		}

		var provider gandi.Provider
		switch internal.DNSProvider {
		case "gandi":
			provider, err = gandiPkg.New()
			if err != nil {
				logrus.Fatal(err)
			}
		default:
			logrus.Fatalf("'%s' is not a supported DNS provider", internal.DNSProvider)
		}

		if err := internal.EnsureDNSTypeFlag(c); err != nil {
			logrus.Fatal(err)
		}

		if err := internal.EnsureDNSNameFlag(c); err != nil {
			logrus.Fatal(err)
		}

		if err := internal.EnsureDNSValueFlag(c); err != nil {
			logrus.Fatal(err)
		}

		record := libdns.Record{
			Type:  internal.DNSType,
			Name:  internal.DNSName,
			Value: internal.DNSValue,
			TTL:   time.Duration(internal.DNSTTL),
		}

		if internal.DNSType == "MX" || internal.DNSType == "SRV" || internal.DNSType == "URI" {
			record.Priority = internal.DNSPriority
		}

		records, err := provider.GetRecords(c.Context, zone)
		if err != nil {
			logrus.Fatal(err)
		}

		for _, existingRecord := range records {
			if existingRecord.Type == record.Type &&
				existingRecord.Name == record.Name &&
				existingRecord.Value == record.Value {
				logrus.Fatal("provider library reports that this record already exists?")
			}
		}

		createdRecords, err := provider.SetRecords(
			c.Context,
			zone,
			[]libdns.Record{record},
		)

		if len(createdRecords) == 0 {
			logrus.Fatal("provider library reports that no record was created?")
		}

		createdRecord := createdRecords[0]

		tableCol := []string{"type", "name", "value", "TTL", "priority"}
		table := abraFormatter.CreateTable(tableCol)

		value := createdRecord.Value
		if len(createdRecord.Value) > 30 {
			value = fmt.Sprintf("%s...", createdRecord.Value[:30])
		}

		table.Append([]string{
			createdRecord.Type,
			createdRecord.Name,
			value,
			createdRecord.TTL.String(),
			strconv.Itoa(createdRecord.Priority),
		})

		table.Render()

		logrus.Info("record created")

		return nil
	},
}
